/***************************************************************************
 *   Copyright (C) 2004 by Murray Evans                                    *
 *   m.evans@rdg.ac.uk                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Mesh.h"
#include "Collision.h"
#include "DynArray.h"

#define IGNORE_X 1
#define IGNORE_Y 2
#define IGNORE_Z 3



// check for a collision with a sphere (point and radius) and a mesh in mesh local space
bool CMesh::CollideMeshSphere(CVec4f in_SphOrigin_v4f, float in_SphRadius_f, SMeshCollision *out_Collision)
{
	// want to build up a list of faces in the mesh for which a collision takes place.
	CDynArray<SMeshCollision>	recordedCollisions;

	// could definately do some form of oct/quad treeing here to cut down on how much
	// of the mesh is checked against... Particularly as this will doubtless be used 
	// for wheel-terrain checks so needs to be quick...

	for(int fc_i = 0; fc_i < faces.Size(); fc_i++)
	{
		// get D value of the plane of the face (Ax+By+Cz+D=0)
		CVec4f pop;
		pop = points[faces[fc_i]->p[0]];
		CVec4f fn;
		fn = fnormals[faces[fc_i]->fn];
		fn = fn * -1;
      	float  D_f;
		D_f = -(fn.X() * pop.X()) - (fn.Y() * pop.Y()) - (fn.Z() * pop.Z());
		CVec4f rad;

		// does rad need to be in opposite direction to fn??
		rad.SetXYZW(fn.X() * -in_SphRadius_f, fn.Y() * -in_SphRadius_f, fn.Z() * -in_SphRadius_f, 0);

		float top, bot;
        top = -D_f - fn.Dot(in_SphOrigin_v4f);
        bot = fn.Dot(rad);

		float sc_f;
		if(bot!=0)              // if it does then the line SP->EP lies along the plane.
		{
			sc_f = top/bot;       // sc : scalar value (point_on_line = SP+(sc*dir)) where point_on_line is IP of line SP->EP and plane
			//fprintf(stderr, "scf: %f\n", sc_f);
			if((sc_f < 1) && (sc_f > 0))
			{
				// hence, intersection point of radius and plane is 
				CVec4f ip;
				ip = in_SphOrigin_v4f + (rad * sc_f);

				// but are we intersecting the face, or just the plane of the face??

				// let's be vague and check with a bounding box of the face
				// figure out most significant axis of plane.
				float su_f, bu_f, sv_f, bv_f, ipu_f, ipv_f;
				int ignore_i;
				float m_f = max(max(fn.X(), fn.Y()), fn.Z());
				if(m_f == fn.X())
				{
					ignore_i = IGNORE_X;
				}
				else if(m_f == fn.Y())
				{
					ignore_i = IGNORE_Y;
				}
				else
				{
					ignore_i = IGNORE_Z;
				}

				
				if(ignore_i == IGNORE_X)
				{
					su_f = SmallestY(fc_i);
					bu_f = BiggestY(fc_i);
					sv_f = SmallestZ(fc_i);
					bv_f = BiggestZ(fc_i);
					ipu_f = ip.Y();
					ipv_f = ip.Z();
				}
				else if(ignore_i == IGNORE_Y)
				{
					su_f = SmallestX(fc_i);
					bu_f = BiggestX(fc_i);
					sv_f = SmallestZ(fc_i);
					bv_f = BiggestZ(fc_i);
					ipu_f = ip.X();
					ipv_f = ip.Z();
				}
				else if(ignore_i == IGNORE_Z)
				{
					su_f = SmallestX(fc_i);
					bu_f = BiggestX(fc_i);
					sv_f = SmallestY(fc_i);
					bv_f = BiggestY(fc_i);
					ipu_f = ip.X();
					ipv_f = ip.Y();
				}
				
				if((ipu_f > su_f) && (ipu_f < bu_f))
					if ((ipv_f > sv_f) && (ipv_f < bv_f))
					{
						// collision with bounding box of face.
						// be more specific??

						// record collision
						SMeshCollision temp_mcol;
						temp_mcol.face_i  = fc_i;
						temp_mcol.ip_f    = sc_f;
						temp_mcol.ip_v4f  = ip;
						temp_mcol.rad_v4f = rad;
						recordedCollisions.AddToEnd(temp_mcol);
					}		
			}
		}		
	}

	// want the recorded collision closest to origin of sphere
	float check_f =1;
	int which_i = -1;
	for(int c_i = 0; c_i < recordedCollisions.Size(); c_i++)
	{
		if(recordedCollisions[c_i].ip_f < check_f)
		{
			which_i = c_i;
		}
	}
	if(which_i != -1) 
	{
		*out_Collision = recordedCollisions[which_i];
		return true;
	}

	return false;
}

// calculate for a given face the smallest/largest X,Y,Z values
inline float CMesh::SmallestX(int in_Face_i)
{
	float out_SmallX_f = points[faces[in_Face_i]->p[0]].X();
	for(int pc_i = 1; pc_i < faces[in_Face_i]->p.Size(); pc_i++)
	{
		float temp_f = points[faces[in_Face_i]->p[pc_i]].X();
		if(temp_f < out_SmallX_f) out_SmallX_f = temp_f;
	}
	return out_SmallX_f;
}
inline float CMesh::SmallestY(int in_Face_i)
{
	float out_SmallY_f = points[faces[in_Face_i]->p[0]].Y();
	for(int pc_i = 1; pc_i < faces[in_Face_i]->p.Size(); pc_i++)
	{
		float temp_f = points[faces[in_Face_i]->p[pc_i]].Y();
		if(temp_f < out_SmallY_f) out_SmallY_f = temp_f;
	}
	return out_SmallY_f;
}
inline float CMesh::SmallestZ(int in_Face_i)
{
	float out_SmallZ_f = points[faces[in_Face_i]->p[0]].Z();
	for(int pc_i = 1; pc_i < faces[in_Face_i]->p.Size(); pc_i++)
	{
		float temp_f = points[faces[in_Face_i]->p[pc_i]].Z();
		if(temp_f < out_SmallZ_f) out_SmallZ_f = temp_f;
	}
	return out_SmallZ_f;
}
inline float CMesh::BiggestX(int in_Face_i)
{
	float out_BigX_f = points[faces[in_Face_i]->p[0]].X();
	for(int pc_i = 1; pc_i < faces[in_Face_i]->p.Size(); pc_i++)
	{
		float temp_f = points[faces[in_Face_i]->p[pc_i]].X();
		if(temp_f > out_BigX_f) out_BigX_f = temp_f;
	}
	return out_BigX_f;
}
inline float CMesh::BiggestY(int in_Face_i)
{
	float out_BigY_f = points[faces[in_Face_i]->p[0]].Y();
	for(int pc_i = 1; pc_i < faces[in_Face_i]->p.Size(); pc_i++)
	{
		float temp_f = points[faces[in_Face_i]->p[pc_i]].Y();
		if(temp_f > out_BigY_f) out_BigY_f = temp_f;
	}
	return out_BigY_f;
}
inline float CMesh::BiggestZ(int in_Face_i)
{
	float out_BigZ_f = points[faces[in_Face_i]->p[0]].Z();
	for(int pc_i = 1; pc_i < faces[in_Face_i]->p.Size(); pc_i++)
	{
		float temp_f = points[faces[in_Face_i]->p[pc_i]].Z();
		if(temp_f > out_BigZ_f) out_BigZ_f = temp_f;
	}
	return out_BigZ_f;
}

#undef IGNORE_X
#undef IGNORE_Y
#undef IGNORE_Z